
from argparse import ArgumentParser
import matplotlib.pyplot as plt

import torch

# Local imports
from localperf.core.measuring import measure_time, deal_with_results
from localperf.core.utils import create_dir, remove_file
from localperf.core.compute import compute, treat_batch
from localperf.core.config import default_log_n_data_torch, default_n_measures_torch, n_neurons_torch_model
plt.show()
class Net(torch.nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.fc1 = torch.nn.Linear(n_neurons_torch_model, n_neurons_torch_model)
        self.fc2 = torch.nn.Linear(n_neurons_torch_model, n_neurons_torch_model)
        self.fc3 = torch.nn.Linear(n_neurons_torch_model, n_neurons_torch_model)
        self.fc4 = torch.nn.Linear(n_neurons_torch_model, 10)
    def forward(self, x):
        x = self.fc1(x)
        x = self.fc2(x)
        x = self.fc3(x)
        x = self.fc4(x)
        return x

def get_model(device : torch.device):
    """Create a model and move it to the device.

    Args:
        device (torch.device): the device on which the model will be moved.

    Returns:
        torch.nn.Module: the torch model.
    """
    model = Net()
    model.to(device = device)
    batch = torch.rand(size = (2, n_neurons_torch_model), device=device)
    treat_batch(model=model, batch=batch, device=device)
    return model
    
    
    
if __name__ == "__main__":
    # Parser
    parser = ArgumentParser()
    parser.add_argument("--log_dir", type=str, default=None, help="Directory where to save the logs. No saving by default")
    parser.add_argument("--plot", action="store_true", default=False, help="Plot the results. No plotting by default")
    parser.add_argument("--image_dir", type=str, default=None, help="Directory where to save the images. No saving by default")
    parser.add_argument("--no-progress", action="store_true", help="Hide the progress bar.")
    
    parser.add_argument("--log_n_data", type=int, default=default_log_n_data_torch, help=f"Value (in log scale) of the maximum n_data to be tested. Default: {default_log_n_data_torch} (10^{default_log_n_data_torch})")
    parser.add_argument("--n_measures", type=int, default=default_n_measures_torch, help=f"Number of measures to be made for each n_data. Default: {default_n_measures_torch}")
    parser.add_argument("--n_measures_gpu", type=int, default=None, help=f"Number of measures to be made for each n_data for the GPU. Default: same as --n_measures")                    
    args = parser.parse_args()

    image_dir = args.image_dir
    log_dir = args.log_dir
    do_plot = args.plot
    log_n_data_max = args.log_n_data
    n_measures = args.n_measures
    n_measures_gpu = args.n_measures_gpu if args.n_measures_gpu is not None else n_measures
    show_progress_bar = not args.no_progress



    # Setup
    print(
f"===== GPU measurement ===== \n\
Speed will be measured for data in range [1, 10^{log_n_data_max}] \n\
and for {n_measures} measures for each data. \n\
===========================\n\
        ")
    list_n_data = [10**k for k in range(0, log_n_data_max + 1)]
    log_filename = log_dir + "/gpu_torch.txt" if log_dir is not None else None
    image_filename = image_dir + "/gpu_torch.png" if image_dir is not None else None
    create_dir(log_dir)
    create_dir(image_dir)
    remove_file(log_filename)
    
    
    
    # Measure CPU_only torch performance
    device = "cpu"
    model = get_model(device = device)
    title = f"Torch with {device}"
    print(title)
    
    def cpu_only_torch_compute(n_data : int):
        batch = torch.rand(size = (n_data, n_neurons_torch_model), device = "cpu")
        treat_batch(model=model, batch=batch, device="cpu")
        
        
    list_mean_time_cpu, list_std_time = measure_time(
        func = cpu_only_torch_compute, 
        list_inputs = list_n_data, 
        n_measures = n_measures,
        show_progress_bar = show_progress_bar,
        )
            
    deal_with_results(
        list_inputs=list_n_data,
        list_mean_time=list_mean_time_cpu,
        list_std_time=list_std_time,
        do_print=True,
        do_plot=not torch.cuda.is_available(),
        log_filename=log_filename,
        image_filename=image_filename,
        title = title,
    )        
    
    
    # Measure GPU torch performance
    if not torch.cuda.is_available():
        print("No GPU available. Skipping GPU torch performance measurement")
    else:
        device = torch.device("cuda")
        model = get_model(device = device)
        title = f"Torch with {device}"
        print(title)
        
        def gpu_torch_compute(n_data : int):
            batch = torch.rand(size = (n_data, n_neurons_torch_model), device = device)
            treat_batch(model=model, batch=batch, device=device)
            
            
        list_mean_time_gpu, list_std_time = measure_time(
            func = gpu_torch_compute, 
            list_inputs = list_n_data, 
            n_measures = n_measures_gpu,
            show_progress_bar = show_progress_bar,
            )
        list_speed_up = [list_mean_time_cpu[i] / list_mean_time_gpu[i] for i in range(len(list_mean_time_cpu))]
                
        deal_with_results(
            list_inputs=list_n_data,
            list_mean_time=list_mean_time_gpu,
            list_std_time=list_std_time,
            list_speed_up=list_speed_up,
            do_print=True,
            do_plot=do_plot,
            log_filename=log_filename,
            image_filename=image_filename,
            title = title,
        )


"""This module contains functions aimed at simulating intensive computations.
"""

from argparse import ArgumentParser
from time import time, perf_counter
import numpy as np
import matplotlib.pyplot as plt
from typing import Callable, List, Any, Tuple

from localperf.core.config import data_size

def treat_one_data():
    """Compute the sum of the first n integers.
    This is the most basic function to compute. Its computation correspond to the treatment of a single data.
    Complexity is O(1) = BaseComplexity. This is the base unit of complexity in our package."""
    sum_x = 0
    for i in range(data_size):
        sum_x += i
    return sum_x

def compute(n_data : int):
    """Compute iteratively.
    Complexity is O(n_data) = BaseComplexity * n_data"""
    for _ in range(n_data):
        treat_one_data()


def treat_batch(model : "torch.nn.Module", batch : "torch.Tensor", device : "torch.cuda.device"):
    """Realize a forward pass on a batch.

    Args:
        model (torch.nn.Module): a neural network
        batch (torch.Tensor): the batch to be treated
        device (torch.device): the device on which the computation is performed
    """
    # print("Device:", torch.cuda.current_device())
    model(batch)


from argparse import ArgumentParser
from time import time, perf_counter
import numpy as np
import matplotlib.pyplot as plt
from typing import Callable, List, Any, Tuple
from tqdm import tqdm

from localperf.core.utils import build_plt_figure, get_results_as_string
from localperf.core.utils import create_dir, remove_file

def measure_time(
        func : Callable, 
        list_inputs : List[Any], 
        n_measures : int = 10,
        show_progress_bar : bool = False,
        ) -> Tuple[List[float], List[float]]:
    """Measure the mean and std of the time taken by a function, for each input in list_input.

    Args:
        func (Callable): the function we want to measure
        list_input (List[Any]): the list of inputs for the function func
        n_measures (int, optional): The number of measures that will be made for evaluating the mean and std. Defaults to 10.
        show_progress_bar (bool, optional): Whether to show a progress bar. Defaults to False.
        
    Returns:
        Tuple[List[float], List[float]]: The list of mean and std of the time taken by the function func, for each input in list_input.
    """
    list_mean_time = []
    list_std_time = []
    for x_input in list_inputs:
        list_time = []
        
        if show_progress_bar:
            iterable = tqdm(range(n_measures), desc=f"Measuring time for {x_input} data")
        else:
            iterable = range(n_measures)
        
        for _ in iterable:
            t_start = perf_counter()
            func(x_input)
            t_end = perf_counter()
            list_time.append(t_end - t_start)
        list_mean_time.append(np.mean(list_time))
        list_std_time.append(np.std(list_time))
    return list_mean_time, list_std_time
    
    

def deal_with_results(
        list_inputs : List[Any], 
        list_mean_time : List[float], 
        list_std_time : List[float], 
        list_speed_up : List[float] = None,
        do_print : bool = False, do_plot : bool = False, 
        log_filename : str = None, image_filename : str = None,
        title : str = None,
        ) -> None:
    """Deals with the results obtained, by (eventually) printing, plotting and saving logs and images.

    Args:
        list_inputs (List[Any]): the list of inputs for the function func. If can't be print, you can put None.
        list_mean_time (List[float]): the list of mean of the time taken by the function func, for each input in list_input.
        list_std_time (List[float]): the list of std of the time taken by the function func, for each input in list_input.
        list_speed_up (List[float], optional): the list of speed up obtained. Defaults to None.
        do_print (bool): whether to print the results on the terminal.
        do_plot (bool): whether to plot the results.
        log_filename (str): filename for the log file.
        image_filename (str): filename for the image file.
    """
    if list_inputs is None:
        list_inputs = ["-" for _ in list_mean_time]
        
    if do_print:
        string = string = get_results_as_string(list_inputs, list_mean_time, list_std_time, list_speed_up = list_speed_up)
        print(string)
    
    if log_filename is not None:
        string = get_results_as_string(list_inputs, list_mean_time, list_std_time, list_speed_up = list_speed_up)
        if title is not None:
            string = title + "\n" + string + "\n"
        with open(log_filename, "a") as f:
            f.write(string)
    
    build_plt_figure(list_inputs, list_mean_time, list_std_time, title)
    if image_filename is not None:
        plt.savefig(image_filename)  
    if do_plot:        
        plt.show()
        
    



from argparse import ArgumentParser
from time import time, perf_counter
import numpy as np
import matplotlib.pyplot as plt
from typing import Callable, List, Any, Tuple
from joblib import Parallel, delayed

# Local imports
from localperf.core.measuring import measure_time, deal_with_results
from localperf.core.utils import create_dir, remove_file
from localperf.core.compute import treat_one_data
from localperf.core.config import default_log_n_data_mp, default_log2_n_process_mp, default_n_measures_mp


            
            
if __name__ == "__main__":
    # Parser
    parser = ArgumentParser()
    parser.add_argument("--log_dir", type=str, default=None, help="Directory where to save the logs. No saving by default")
    parser.add_argument("--plot", action="store_true", default=False, help="Plot the results. No plotting by default")
    parser.add_argument("--image_dir", type=str, default=None, help="Directory where to save the images. No saving by default")
    parser.add_argument("--no-progress", action="store_true", help="Hide the progress bar.")
    
    parser.add_argument("--log_n_data", type=int, default=default_log_n_data_mp, help=f"Value (in log10 scale) of the maximum n_data to be tested. Default: {default_log_n_data_mp} (10^{default_log_n_data_mp} data max)")
    parser.add_argument("--log2_n_process", type=int, default=default_log2_n_process_mp, help=f"Value (in log2 scale) of the maximum n_process to be tested. Default: {default_log2_n_process_mp} ({2**default_log2_n_process_mp} process max)")
    parser.add_argument("--n_measures", type=int, default=default_n_measures_mp, help=f"Number of measures to be made for each n_data. Default: {default_n_measures_mp}")
    
    args = parser.parse_args()
    
    image_dir = args.image_dir
    log_dir = args.log_dir
    do_plot = args.plot
    log_n_data_max = args.log_n_data
    log2_n_process_max = args.log2_n_process
    n_measures = args.n_measures    
    show_progress_bar = not args.no_progress

    

    # Setup
    print(
f"===== Parallelization speed-up measurement ===== \n\
Speed up with parallelization with Joblib will be measured \n\
For data in range [1, 10^{log_n_data_max}] \n\
For n_process in range [1, 2^{log2_n_process_max}] = [1, {2**log2_n_process_max}] \n\
With {n_measures} measures for each data. \n\
================================================\n\
        ")
    list_n_data = [10**k for k in range(0, log_n_data_max + 1)]
    list_n_process = [2**k for k in range(0, log2_n_process_max + 1)]
    log_filename = log_dir + "/multiprocessing.txt" if log_dir is not None else None
    image_filename = image_dir + "/multiprocessing.png" if image_dir is not None else None
    create_dir(log_dir)
    create_dir(image_dir)
    remove_file(log_filename)
    
    
    
    # Measure with no parallelization
    title = f"No parallelization"
    print(title)
        
    def for_loop_computing(n_data: int):
        "Compute data sequentially."
        for _ in range(n_data):
            treat_one_data()
            
    list_mean_time_no_parallelization, list_std_time = measure_time(
        func = for_loop_computing, 
        list_inputs = list_n_data, 
        n_measures = n_measures,
        show_progress_bar = show_progress_bar,
        )

    deal_with_results(
        list_inputs=list_n_data,
        list_mean_time=list_mean_time_no_parallelization,
        list_std_time=list_std_time,
        do_print=True,
        do_plot=False,
        log_filename=log_filename,
        image_filename=image_filename,
        title=title,
    ) 
        
        
        
    for n_process in list_n_process:
        
        # Measure with joblib with n_process processes
        title=f"Joblib parallel° (n_process={n_process})"
        print(title)
        
        def parallel_computing_with_joblib(n_data: int):
            """Compute data in parallel with joblib."""
            Parallel(n_jobs=n_process)(delayed(treat_one_data)() for _ in range(n_data))
    
        list_mean_time, list_std_time = measure_time(
            func = parallel_computing_with_joblib, 
            list_inputs = list_n_data, 
            n_measures = n_measures,
            show_progress_bar = show_progress_bar,
            )
        list_speed_up = [list_mean_time_no_parallelization[i] / list_mean_time[i] for i in range(len(list_mean_time))]
        
        deal_with_results(
            list_inputs=list_n_data,
            list_mean_time=list_mean_time,
            list_std_time=list_std_time,
            list_speed_up=list_speed_up,
            do_print=True,
            do_plot=do_plot if n_process == list_n_process[-1] else False,
            log_filename=log_filename,
            image_filename=image_filename,
            title=title,
        )        


# Common config
data_size = 1000

# CPU config
default_n_measures_cpu = 10
default_log_n_data_cpu = 4

# Multiprocessing config
default_n_measures_mp = 10
default_log_n_data_mp = 4
default_log2_n_process_mp = 3

# Torch config
default_n_measures_torch = 10
default_log_n_data_torch = 6
n_neurons_torch_model = 10



from argparse import ArgumentParser
from time import time, perf_counter
import numpy as np
import matplotlib.pyplot as plt
from typing import Callable, List, Any, Tuple

# Local imports
from localperf.core.measuring import measure_time, deal_with_results
from localperf.core.utils import create_dir, remove_file
from localperf.core.compute import compute
from localperf.core.config import default_log_n_data_cpu, default_n_measures_cpu            

      
if __name__ == "__main__":
    # Parser
    parser = ArgumentParser()
    parser.add_argument("--log_dir", type=str, default=None, help="Directory where to save the logs. No saving by default")
    parser.add_argument("--plot", action="store_true", default=False, help="Plot the results. No plotting by default")
    parser.add_argument("--image_dir", type=str, default=None, help="Directory where to save the images. No saving by default")
    parser.add_argument("--no-progress", action="store_true", help="Hide the progress bar.")
    
    parser.add_argument("--log_n_data", type=int, default=default_log_n_data_cpu, help=f"Value (in log scale) of the maximum n_data to be tested. Default: {default_log_n_data_cpu} (10^{default_log_n_data_cpu})")
    parser.add_argument("--n_measures", type=int, default=default_n_measures_cpu, help=f"Number of measures to be made for each n_data. Default: {default_n_measures_cpu}")
                        
    args = parser.parse_args()

    image_dir = args.image_dir
    log_dir = args.log_dir
    do_plot = args.plot
    log_n_data_max = args.log_n_data
    n_measures = args.n_measures
    show_progress_bar = not args.no_progress


    # Setup
    print(
f"===== CPU measurement ===== \n\
CPU speed will be measured for data in range [1, 10^{log_n_data_max}] \n\
and for {n_measures} measures for each data. \n\
===========================\n\
        ")
    list_n_data = [10**k for k in range(0, log_n_data_max + 1)]
    log_filename = log_dir + "/cpu.txt" if log_dir is not None else None
    image_filename = image_dir + "/cpu.png" if image_dir is not None else None
    create_dir(log_dir)
    create_dir(image_dir)
    remove_file(log_filename)
    
    
    
    # Measure CPU time
    list_mean_time, list_std_time = measure_time(
        func = compute, 
        list_inputs = list_n_data, 
        n_measures = n_measures,
        show_progress_bar = show_progress_bar,
        )
        
    deal_with_results(
        list_inputs=list_n_data,
        list_mean_time=list_mean_time,
        list_std_time=list_std_time,
        do_print=True,
        do_plot=do_plot,
        log_filename=log_dir + "/cpu.txt" if log_dir is not None else None,
        image_filename=image_dir + "/cpu.png" if image_dir is not None else None,
        title = "CPU",
    )        


from setuptools import setup, find_namespace_packages

setup(
    name="localperf",
    url="https://github.com/tboulet/Local-Python-Performance", 
    author="Timothé Boulet",
    author_email="timothe.boulet0@gmail.com",
    
    packages=find_namespace_packages(),
    requires=[
        "numpy",
        "joblib",
        "matplotlib",
        "tqdm",
    ],
    version="1.2",
    license="MIT",
    description="Measure of python performance in local.",
    long_description=open('README.md').read(),      
    long_description_content_type="text/markdown",
)


from argparse import ArgumentParser
from time import time, perf_counter
import numpy as np
import matplotlib.pyplot as plt
from typing import Callable, List, Any, Tuple
import os


def build_plt_figure(list_inputs, list_mean_time, list_std_time, title : str):
    """Build the matplotlib figure for the CPU performance."""
    plt.figure("Local Python Performance")
    plt.errorbar(list_inputs, list_mean_time, yerr=list_std_time, label=title)
    plt.xlabel("n_data")
    plt.ylabel("Time (s)")
    plt.legend(loc = "upper left")
    plt.xscale("log")
    plt.yscale("log")
    
def get_results_as_string(list_inputs, list_mean_time, list_std_time, list_speed_up = None):
    """Generate results as string."""
    if list_speed_up is not None:
        string = "n_data\tmean_time\tstd_time\tspeed_up (time taken divided by basic_python_time)\n"
        for x_input, mean_time, std_time, speed_up in zip(list_inputs, list_mean_time, list_std_time, list_speed_up):
            string += f"{x_input}\t{mean_time:.2e}\t{std_time:.2e}\t{speed_up}\n"
    else:
        string = "n_data\tmean_time\tstd_time\n"
        for x_input, mean_time, std_time in zip(list_inputs, list_mean_time, list_std_time):
            string += f"{x_input}\t{mean_time:.2e}\t{std_time:.2e}\n"
    return string

def create_dir(directory : str = None):
    """Create a directory if it does not exist."""
    if directory is not None and not os.path.exists(directory):
        os.makedirs(directory)
        
def remove_file(filename : str = None):
    """Remove a file if it exists."""
    if filename is not None and os.path.exists(filename):
        os.remove(filename)